#include	<X11/X.h>
#include	<X11/Xlib.h>
#include	<X11/Xutil.h>
#include	<X11/Xos.h>
#include	<X11/Xresource.h>
#include	<X11/Xatom.h>

#include	<stdio.h>
#include	<stdlib.h>

#include	"wm.h"

/*
 *    Dispatcher for main event loop.
 */
typedef struct Disp Disp;
struct Disp {
	int	type;
	void	(*handler)(XEvent *);
};

static void expose(XEvent *);
static void buttonpress(XEvent *);
static void buttonrelease(XEvent *);
static void maprequest(XEvent *);
static void configurereq(XEvent *);
static void unmap(XEvent *);
static void destroy(XEvent *);
static void clientmessage(XEvent *);
static void colormap(XEvent *);
static void property(XEvent *);
static void reparent(XEvent *);
static void enter(XEvent *);
static void expose(XEvent *);
static void motionnotify(XEvent *);

static void reshaping_motionnotify(XEvent *);

static Disp disps[] =
{
	{Expose, expose},
	{MotionNotify, motionnotify},
	{ButtonPress, buttonpress},
	{ButtonRelease, buttonrelease},
	{MapRequest, maprequest},
	{ConfigureRequest, configurereq},
	{UnmapNotify, unmap},
	{DestroyNotify, destroy},
	{ClientMessage, clientmessage},
	{ColormapNotify, colormap},
	{PropertyNotify, property},
	{ReparentNotify, reparent},
	{EnterNotify, enter},
	{CirculateRequest, 0},
	{ConfigureNotify, 0},
	{CreateNotify, 0},
	{GravityNotify, 0},
	{MapNotify, 0},
	{MappingNotify, 0},
	{SelectionClear, 0},
	{SelectionNotify, 0},
	{SelectionRequest, 0},
};

extern void
dispatch(XEvent *ev)
{
	Disp *p;

#ifdef ZOE_KANARI
	fprintf(stderr, "<");
	switch (ev->type) {
	case ButtonPress:
		fprintf(stderr, "ButtonPress"); break;
	case ButtonRelease:
		fprintf(stderr, "ButtonRelease"); break;
	case MapRequest:
		fprintf(stderr, "MapRequest"); break;
	case ConfigureRequest:
		fprintf(stderr, "ConfigureRequest"); break;
	case UnmapNotify:
		fprintf(stderr, "UnmapNotify"); break;
	case DestroyNotify:
		fprintf(stderr, "DestroyNotify"); break;
	case ClientMessage:
		fprintf(stderr, "ClientMessage"); break;
	case ColormapNotify:
		fprintf(stderr, "ColourmapNotify"); break;
	case PropertyNotify:
		fprintf(stderr, "PropertyNotify"); break;
	case ReparentNotify:
		fprintf(stderr, "ReparentNotify"); break;
	case EnterNotify:
		fprintf(stderr, "EnterNotify"); break;
	case Expose:
	case MotionNotify:
		/* These are too frequent and uninteresting to be displayed. */
		break;
	case CirculateRequest:
		fprintf(stderr, "CirculateRequest"); break;
	case ConfigureNotify:
		fprintf(stderr, "ConfigureNotify"); break;
	case CreateNotify:
		fprintf(stderr, "CreateNotify"); break;
	case GravityNotify:
		fprintf(stderr, "GravityNotify"); break;
	case MapNotify:
		fprintf(stderr, "MapNotify"); break;
	case MappingNotify:
		fprintf(stderr, "MappingNotify"); break;
	case SelectionClear:
		fprintf(stderr, "SelectionClear"); break;
	case SelectionNotify:
		fprintf(stderr, "SelectionNotify"); break;
	case SelectionRequest:
		fprintf(stderr, "SelectionRequest"); break;
	default:
		fprintf(stderr, "unknown event type %i", ev->type);
	}
	fprintf(stderr, ">");
#endif

	for (p = disps; p < disps + sizeof disps / sizeof disps[0]; p++) {
		if (p->type == ev->type) {
			if (p->handler != 0)
				p->handler(ev);
			return;
		}
	}

	if (!shapeevent(ev))
		fprintf(stderr, "lwm: unknown event %d\n", ev->type);
}

static void
expose(XEvent *ev)
{
	Client *c;
	Window w;	/* Window the expose event is for. */

	/*
	 * Only handle the last in a group of Expose events.
	 */
	if (ev->xexpose.count != 0) return;

	w = ev->xexpose.window;

	/*
	 * We don't draw on the root window so that people can have
	 * their favourite Spice Girls backdrop...
	 */
	if (w == root) return;

	/*
	 * Decide what needs redrawing: window frame or menu?
	 */
	if (w == popup)
		menu_expose();
	else {
		c = Client_Get(w);
		if (c != 0) {
			Client_DrawBorder(c, c == current);
		}
	}
}

static int
titleheightc(Client *c)
{
	return (c->accepts_focus) ? titleheight() : 0;
}

static void
buttonpress(XEvent *ev)
{
	Client *c;
	XButtonEvent *e = &ev->xbutton;
	int quarter;

	/* If we're getting it already, we're not in the market for more. */
	if (mode != wm_idle) return;

	c = Client_Get(e->window);
	if (c && (e->window == c->parent)) {
		/*
		 *	Click went to our frame around a client.
		 */

		/*
		 *	The ``box''.
		 */
		if (c->accepts_focus) {
			quarter = (border + titleheight()) / 4;
			if (e->x > border && e->x < (border + 2 * quarter) &&
			    e->y > quarter && e->y < 3 * quarter) {
				Client_Close(c);
				return;
			}
		}
		if (e->button == Button3) {
			hide(c);
			return;
		}
		if (e->button == Button2) {
			Client_Move(c);
			return;
		}
		if (e->button == Button1) {
			/*
			 *	Reshape
			 */

			XMapRaised(dpy, c->parent);

			/* Lasciate ogni speranza voi ch'entrate...  */

			if (e->x <= border && e->y <= border)
				Client_ReshapeEdge(c, ETopLeft);
			else if (e->x >= (c->size.width - border) && e->y <= border)
				Client_ReshapeEdge(c, ETopRight);
			else if (e->x >= (c->size.width - border) &&
						e->y >= (c->size.height + titleheightc(c) - border))
				Client_ReshapeEdge(c, EBottomRight);
			else if (e->x <= border &&
						e->y >= (c->size.height + titleheightc(c) - border))
				Client_ReshapeEdge(c, EBottomLeft);
			else if (e->x > border && e->x < (c->size.width - border) &&
						e->y < (titleheightc(c) + border))
				Client_ReshapeEdge(c, ETop);
			else if (e->x > (c->size.width - border) &&
						e->y > border &&
						e->y < (c->size.height + titleheightc(c) - border))
				Client_ReshapeEdge(c, ERight);
			else if (e->x > border && e->x < (c->size.width - border) &&
						e->y > (c->size.height - border))
				Client_ReshapeEdge(c, EBottom);
			else if (e->x < border &&
						e->y > border &&
						e->y < (c->size.height + titleheightc(c) - border))
				Client_ReshapeEdge(c, ELeft);
			return;
		}
		return;
	}

	/* If it wasn't a root-menu request, we're not interested. */
	if (e->window != root || e->button != Button3)
		return;

	cmapfocus(0);
	menuhit(e);
}

static void
menu_buttonrelease(XEvent *ev)
{
	int x,y;		/* Pointer position. */
	int n;		/* Menu item. */

	/*
	 * Work out which menu item the button was released over.
	 */
	x = ev->xbutton.x;
	y = ev->xbutton.y;
	n = menu_whichitem(x, y);

	/*
	 * Hide the menu until it's needed again.
	 */
	XUnmapWindow(dpy, popup);

	/*
	 * Do the menu thing.
	 */
	if (n == 0)
		shell();
	else
		unhide(n - 1, 1);

	if (current)
		cmapfocus(current);
}

static void
buttonrelease(XEvent *ev)
{
	if (mode == wm_menu_up)
		menu_buttonrelease(ev);

	mode = wm_idle;
}

static void
maprequest(XEvent *ev)
{
	Client *c;
	XMapRequestEvent	*e = &ev->xmaprequest;

	c = Client_Get(e->window);

	if (c == 0 || c->window != e->window) {
		c = Client_Add(e->window);
		if (c == 0)
			return;
	}

	switch (c->state) {
	case WithdrawnState:
		if (c->parent == root) {
			manage(c, 0);
			break;
		}
		XReparentWindow(dpy, c->window, c->parent, border, border);
		XAddToSaveSet(dpy, c->window);
		/*FALLTHROUGH */
	case NormalState:
		XMapRaised(dpy, c->parent);
		XMapWindow(dpy, c->window);
		Client_SetState(c, NormalState);
		break;
	case IconicState:
		unhidec(c, 1);
		break;
	}
}

static void
unmap(XEvent *ev)
{
	Client *c;
	XUnmapEvent *e = &ev->xunmap;

	c = Client_Get(e->window);
	if (c == 0) return;

	/*
	 * In the description of the ReparentWindow request we read: "If the window
	 * is mapped, an UnmapWindow request is performed automatically first". This
	 * might seem stupid, but it's the way it is. While a reparenting is pending
	 * we ignore UnmapWindow requests.
	 */
	if (c->internal_state == IPendingReparenting) {
		c->internal_state = INormal;
		return;
	}

	/* "This time it's the real thing." */

	if (c->state == IconicState && e->send_event) {
		unhidec(c, 0);
	}

	withdraw(c);

	/*
	 * In response to having a window closed by the user, an application
	 * may choose to simply unmap the window. We shouldn't in this case
	 * bear a grudge against the window any longer, and internal_state should
	 * not be IPendingDeletion.
	 */
	c->internal_state = INormal;
}

static void
configurereq(XEvent *ev)
{
	XWindowChanges wc;
	Client *c;
	XConfigureRequestEvent *e = &ev->xconfigurerequest;

	c = Client_Get(e->window);
	if (c) {
		if (e->value_mask & CWX)
			c->size.x = e->x;
		if (e->value_mask & CWY)
			c->size.y = e->y;
		if (e->value_mask & CWWidth)
			c->size.width = e->width + 2 * border;
		if (e->value_mask & CWHeight)
			c->size.height = e->height + 2 * border;
		if (e->value_mask & CWBorderWidth)
			c->border = e->border_width;

		if (c->parent != root && c->window == e->window) {
			wc.x = c->size.x - border;
			wc.y = c->size.y - border;
			wc.width = c->size.width;
			wc.height = c->size.height;
			if (c == current)
				wc.height += titleheight();
			wc.border_width = 1;
			wc.sibling = e->above;
			wc.stack_mode = e->detail;
			XConfigureWindow(dpy, e->parent, e->value_mask, &wc);
			SendConfigureNotify(c);
		}
	}
	if (c && (c->internal_state == INormal)) {
		wc.x = border;
		wc.y = border;
	} else {
		wc.x = e->x;
		wc.y = e->y;
	}
	wc.width = e->width;
	wc.height = e->height;
	wc.border_width = 0;
	wc.sibling = e->above;
	wc.stack_mode = e->detail;
	e->value_mask |= CWBorderWidth;

	XConfigureWindow(dpy, e->window, e->value_mask, &wc);
}

static void
destroy(XEvent *ev)
{
	Client	*c;
	Window	w = ev->xdestroywindow.window;

	c = Client_Get(w);
	if (c == 0)
		return;

	Client_Remove(c);
}

static void
clientmessage(XEvent *ev)
{
	Client	*c;
	XClientMessageEvent	*e = &ev->xclient;

	if (e->message_type == wm_change_state) {
		c = Client_Get(e->window);
		if (c == 0) return;
		if (e->format == 32 && e->data.l[0] == IconicState && normal(c)) hide(c);
		return;
	}
}

static void
colormap(XEvent *ev)
{
	Client	*c;
	XColormapEvent	*e = &ev->xcolormap;

	if (e->new) {
		c = Client_Get(e->window);
		if (c) {
			c->cmap = e->colormap;
			if (c == current)
				cmapfocus(c);
		} else
			Client_ColourMap(ev);
	}
}

static void
property(XEvent *ev)
{
	Client	*c;
	XPropertyEvent	*e = &ev->xproperty;

	c = Client_Get(e->window);
	if (c == 0)
		return;

	switch (e->atom) {
	case XA_WM_NAME:
		getWindowName(c);
		return;
	case XA_WM_TRANSIENT_FOR:
		gettrans(c);
		return;
	default:
		if (e->atom == wm_colormaps) {
			getcmaps(c);
			if (c == current)
				cmapfocus(c);
		}
	}
}

static void
reparent(XEvent *ev)
{
	Client	*c;
	XReparentEvent	*e = &ev->xreparent;

	if (e->event != root || e->override_redirect || e->parent == root)
		return;

	c = Client_Get(e->window);
	if (c != 0 && (c->parent == root || withdrawn(c)))
		Client_Remove(c);
}

static void
enter(XEvent *ev)
{
	Client	*c;

	c = Client_Get(ev->xcrossing.window);
	if (c == 0)
		return;

	if (c != current) {
		/* Grab the server to ward off a nasty race. */
		XGrabServer(dpy);

		XMapWindow(dpy, c->parent);

		/* Take focus away from current holder. */
		if (current)
			setactive(current, 0);

		/* Give focus to new holder. */
		current = c;
		setactive(current, 1);

		/* Ungrab the server, hoping the race danger has passed. */
		XFlush(dpy);
		XUngrabServer(dpy);
	}
}

static void
motionnotify(XEvent *ev)
{
	if (mode == wm_reshaping)
		reshaping_motionnotify(ev);
	else if (mode == wm_menu_up)
		menu_motionnotify(ev);
}

static void
reshaping_motionnotify(XEvent* ev)
{
	int	nx;	/* New x. */
	int	ny;	/* New y. */
	int	ox;	/* Original x. */
	int	oy;	/* Original y. */
	int	ndx;	/* New width. */
	int	ndy;	/* New height. */
	int	odx;	/* Original width. */
	int	ody;	/* Original height. */
	int	pointer_x;
	int	pointer_y;
	int	y_fixup;

	if (mode != wm_reshaping) return;

	y_fixup = (current->accepts_focus) ? titleheight() : 0;

	getmouse(&pointer_x, &pointer_y);

	if (interacting_edge != ENone) {
		nx = ox = current->size.x;
		ny = oy = current->size.y;
		ndx = odx = current->size.width;
		ndy = ody = current->size.height;

		/* Vertical. */
		switch (interacting_edge) {
		case ETop:
		case ETopLeft:
		case ETopRight:
			ndy += (current->size.y - pointer_y);
			ny = pointer_y;
			break;
		case EBottom:
		case EBottomLeft:
		case EBottomRight:
			ndy = pointer_y - current->size.y;
			break;
		default:	break;
		}

		/* Horizontal. */
		switch (interacting_edge) {
		case ERight:
		case ETopRight:
		case EBottomRight:
			ndx = pointer_x - current->size.x;
			break;
		case ELeft:
		case ETopLeft:
		case EBottomLeft:
			ndx += (current->size.x - pointer_x);
			nx = pointer_x;
			break;
		default: break;
		}

		/* Force the suggested values to be OK. */
		Client_MakeSane(current, interacting_edge, &nx, &ny, &ndx, &ndy);

		current->size.x = nx;
		current->size.y = ny;
		current->size.width  = ndx;
		current->size.height = ndy;

		XMoveResizeWindow(dpy, current->parent,
			current->size.x, current->size.y - y_fixup,
			current->size.width, current->size.height + y_fixup);
		if (current->size.width == odx && current->size.height == ody) {
			if (current->size.x != ox || current->size.y != oy)
				SendConfigureNotify(current);
		} else
			XMoveResizeWindow(dpy, current->window, border, border + y_fixup,
			    current->size.width - 2 * border, current->size.height - 2 * border);
	} else {
		nx = pointer_x + start_x;
		ny = pointer_y + start_y;

		Client_MakeSane(current, interacting_edge, &nx, &ny, 0, 0);
		current->size.x = nx;
		current->size.y = ny;
		XMoveWindow(dpy, current->parent,
			current->size.x, current->size.y - y_fixup);
		SendConfigureNotify(current);
	}
}
